'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

exports.default = function (babel) {
  var t = babel.types,
      version = babel.version;

  var babel6 = _semver2.default.satisfies(version, '^6.0.0');

  return {
    name: 'babel-blade', // not required
    inherits: babel6 ? _babelPluginSyntaxJsx2.default : _pluginSyntaxJsx2.default, // for jsx/babel6-7 interop
    visitor: {
      Identifier(path) {
        handleCreateRazor(path, t);
      }
    }
  };
};

exports.handleCreateRazor = handleCreateRazor;

var _semver = require('semver');

var _semver2 = _interopRequireDefault(_semver);

var _pluginSyntaxJsx = require('@babel/plugin-syntax-jsx');

var _pluginSyntaxJsx2 = _interopRequireDefault(_pluginSyntaxJsx);

var _babelPluginSyntaxJsx = require('babel-plugin-syntax-jsx');

var _babelPluginSyntaxJsx2 = _interopRequireDefault(_babelPluginSyntaxJsx);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

// data structure
var _require = require('./dataStructures'),
    RazorData = _require.RazorData; // for jsx/babel6-7 interop


var _require2 = require('./helpers'),
    isObject = _require2.isObject,
    isCallee = _require2.isCallee,
    isCreateQuery = _require2.isCreateQuery,
    isCreateFragment = _require2.isCreateFragment,
    getAssignTarget = _require2.getAssignTarget,
    getObjectPropertyName = _require2.getObjectPropertyName,
    getCalleeArgs = _require2.getCalleeArgs,
    maybeGetSimpleString = _require2.maybeGetSimpleString,
    getSimpleFragmentName = _require2.getSimpleFragmentName;

/****
 *
 * Discussion of babel strategy
 *
 * at the top level we give users createQuery and createFragment methods.
 * when these are called and assigned, whatever identifier they get assigned to becomes a "razor".
 *
 * 1. for each razor
 * 	for each call of the razor that is assigned, that is a blade
 * 		we call parseBlade on the blade
 *
 * 2. parseBlade
 *  only assignment targets are blades
 *    process each reference (variable declarator and standalone) at once
 *
 * 2.1 process reference
 *    lhs = process LHS
 *    rhs process RHS
 *    merge(lhs, rhs)
 *    parseBlade on all assign targets (if any)
 *
 * 2.2 process LHS
 *    get LHS (either id or objectpattern's key)
 *
 * 2.3 process RHS
 *    get RHS list
 *
 * 2.4 merge()
 *    added properties can come from either lhs or rhs. rhs taking precedence
 *      rhs.foreach - call add, return child, call add again
 *      lhs.foreach - ???
 *    if rhs and lhs has no kids, just an assign, just tag the blade and early exit
 *    else, sequence is:
 *      .add from rhs, including args
 *      at the end of rhs, if there is an alias, assign the alias too
 *      if lhs has kids, keep .adding, including alias
 *
 * 3.	once all blades are parsed, compose and insert the graphql query where the razor is referenced
 *
 * LHS cases to handle:
 * - const var1 = DATA
 * - const { var2 } = DATA
 * - const { var3, var4 } = DATA
 * - const { var5: var6  } = DATA
 * - const { var7: { var8: var9} } = DATA
 *
 * interaction cases to handle:
 * - const var1 = DATA.v2
 * - const { var2 } = DATA.v2
 * - const { var3, var4 } = DATA.v2
 * - const { var5: var6  } = DATA.v2
 * - const { var7: { var8: var9} } = DATA.v2
 *
 * RHS cases
 * - const var1 = DATA.v3
 * - const var2 = DATA.v3.var3
 * - const var4 = DATA.v3.var5
 * - const var6 = DATA.v3({ foo: 1, bar: 2})
 * - const var7 = DATA.v3({ foo: 3, bar: 4}).var8
 * - DATA.v3.var8 // no assignment!
 *
 **/

/* eslint-disable complexity */
/* eslint-disable babel/new-cap */

function handleCreateRazor(path, t) {
  if (isCreateQuery(path) || isCreateFragment(path)) {
    // get the identifier and available args
    var identifier = getAssignTarget(path);
    var queryArgs = void 0;
    if (isCallee(path)) {
      queryArgs = getCalleeArgs(path);
    }
    // traverse scope for identifier references
    var refs = path.scope.bindings[identifier].referencePaths;
    // clear the reference
    path.findParent(function (ppath) {
      return ppath.isVariableDeclaration();
    }).remove();
    if (refs.length > 0) {
      var razorID = null;
      var fragmentType = isCreateFragment(path) && maybeGetSimpleString(queryArgs[0]); //getFragmentName(path)
      var queryType = isCreateFragment(path) ? 'fragment' : 'query';
      // console.log({ queryArgs, fragmentType });
      var razorData = new RazorData({
        type: queryType,
        name: isCreateFragment(path) ? t.Identifier(identifier) : identifier,
        fragmentType,
        args: isCreateQuery(path) && queryArgs
      });
      refs.forEach(function (razor) {
        // go through all razors
        if (isCallee(razor)) {
          // we have been activated! time to make a blade!
          razorID = getAssignTarget(razor);
          // clear the reference
          if (razor.container.arguments[0]) razor.parentPath.replaceWith(razor.container.arguments[0]);else razor.parentPath.remove();
          parseBlade(razor, razorID, razorData);
        }
      });

      // insert query
      refs.forEach(function (razor) {
        if (!isObject(razor)) {
          var _razorData$print = razorData.print(),
              stringAccumulator = _razorData$print.stringAccumulator,
              litAccumulator = _razorData$print.litAccumulator;

          var graphqlOutput = t.templateLiteral(stringAccumulator.map(function (str) {
            return t.templateElement({ raw: str });
          }), litAccumulator.map(function (lit) {
            if (lit.isFragment) // we tagged this inside BladeData
              return t.callExpression(lit, [t.stringLiteral(getSimpleFragmentName(lit))]);
            return lit || t.nullLiteral();
          }));
          if (razorData._type === 'fragment') {
            razor.replaceWith(t.arrowFunctionExpression([t.identifier(identifier)], graphqlOutput));
          } else razor.replaceWith(graphqlOutput);
        }
      });
    }
  }
}

function parseBlade(path, id, razorData) {
  var slice = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : 0;

  var refs = path.scope.bindings[id] && path.scope.bindings[id].referencePaths;
  // console.log('parseblade', { refs, id, razorData, path });
  if (slice) refs = refs.slice(slice);
  if (refs && refs.length > 0) {
    // there has been an assignment and it has been used
    refs.forEach(function (blade) {
      processReference(blade, razorData);
      // call parseblade on all LHS
    });
  } else {
    // there has been no assignment or it has not been used
    var bladeID = getObjectPropertyName(path);
    if (bladeID) {
      var propID = getObjectPropertyName(path);
      var child = razorData.add({ name: propID });
      parseBlade(path.parentPath, propID, child);
    }
  }
}

function processReference(blade, razorData) {
  var LHS = void 0,
      RHS = void 0;

  /***** processReference: PROCESS LHS *****/
  // for now LHS processing goes a maximum of one level.
  // will implement multilevel later

  // find declarator parent context
  var ctx = blade.findParent(function (ppath) {
    return ppath.isVariableDeclarator() && ppath.scope.uid === blade.scope.uid;
  });
  if (ctx) {
    // then there is assignment
    // if it is a top level assign, its a plain alias
    if (ctx.node.id.name) {
      LHS = ctx.node.id.name;
    } else {
      LHS = processLHS(ctx);
    }
  }

  /* eslint-disable-next-line */
  function processLHS(ctx) {
    // if it is an object destructure, push and recurse
    var lhs = [];
    var props = ctx.node.id.properties;
    if (props) {
      props.forEach(function (prop) {
        var k = prop.key.name;
        var maybeV = prop.value.properties;
        if (maybeV) {
          // multilayers, recurse
          throw new Error('multilayer destructure not implemented yet');
        } else {
          // lhs and rhs, no multilayer
          lhs.push({
            name: k,
            alias: prop.value.name
          });
        }
      });
      return lhs; // normal terminate
    }
    // shouldnt get here
    // eslint-disable-next-line
    console.log('please examine this', ctx);
  }

  /***** processReference: PROCESS RHS *****/

  // doing this just to read chained objects (eg foo.bar.baz)
  // from left to right instead of
  // babel AST which makes us read from right to left
  RHS = [];
  var RHSVisitor = {
    MemberExpression(childpath) {
      var aliasPath = void 0,
          calleeArguments = void 0;
      if (isCallee(childpath)) {
        // if its a callee, extract its args and push it into RHS
        // will parse out fragments/args/directives later
        calleeArguments = getCalleeArgs(childpath);
        aliasPath = childpath;
      }
      if (childpath.parentKey !== 'arguments')
        // else it will include membexps inside call arguments
        RHS.push({ name: childpath.node.property.name, calleeArguments, aliasPath });
    }
  };
  if (ctx) {
    ctx.traverse(RHSVisitor);
  } else {
    // even if there is no assignment, still need to traverse the kids
    blade.findParent(function (ppath) {
      return !ppath.isMemberExpression();
    }).traverse(RHSVisitor);
  }
  RHS = RHS.reverse(); // annoying bc of how the AST works

  // pretty good place for debugging
  // console.log('bbb', { ctx, blade, LHS, RHS });

  /***** processReference: MERGE LHS AND RHS *****/

  // MERGE RHS FIRST: rhs.foreach - call add, return child, call add again
  var currentData = razorData;
  RHS.forEach(function (_ref) {
    var calleeArguments = _ref.calleeArguments,
        name = _ref.name,
        aliasPath = _ref.aliasPath;


    var args = [];
    var fragments = [];
    var directives = [];

    if (calleeArguments) {
      for (var _iterator = calleeArguments, _isArray = Array.isArray(_iterator), _i = 0, _iterator = _isArray ? _iterator : _iterator[Symbol.iterator]();;) {
        var _ref2;

        if (_isArray) {
          if (_i >= _iterator.length) break;
          _ref2 = _iterator[_i++];
        } else {
          _i = _iterator.next();
          if (_i.done) break;
          _ref2 = _i.value;
        }

        var x = _ref2;

        if (x.type === 'StringLiteral' || x.type === 'TemplateLiteral') {
          // its an arg or a directive; peek at first character to decide
          var peek = x.quasis ? x.quasis[0].value.raw[0] : x.value[0];
          peek === '@' ? directives.push(x) : args.push(x);
        } else {
          // its a fragment
          fragments.push(x);
        }
      }
    }

    currentData = currentData.add({
      name,
      args,
      directives,
      fragments
    });
    if (currentData._args && aliasPath) aliasPath.parentPath.replaceWith(aliasPath);
    if (currentData._alias && aliasPath) aliasPath.node.property.name = currentData._alias;
  });

  // MERGE LHS second: lhs.foreach
  if (typeof LHS === 'string') {
    // simple assignment
    parseBlade(blade.parentPath, LHS, currentData);
  } else if (Array.isArray(LHS)) {
    // there has been a destructuring!
    var LHSpointer = currentData;
    LHS.forEach(function (_ref3) {
      var name = _ref3.name,
          alias = _ref3.alias;

      LHSpointer = currentData.add({ name });
      //const temp = blade.scope.bindings[alias]
      //console.log(alias, blade, temp)
      parseBlade(blade.parentPath, alias, LHSpointer, 1);
    });
  }

  /***** processReference: DONE *****/
}